@@ -1,5 +1,5 @@
 ;;;;
-;;;; Copyright (c) 2008 Zachary Beane, All Rights Reserved
+;;;; Copyright (c) 2008, 2015 Zachary Beane, All Rights Reserved
 ;;;;
 ;;;; Redistribution and use in source and binary forms, with or without
 ;;;; modification, are permitted provided that the following conditions
@@ -89,6 +89,16 @@ constraint."
     (when (plusp (length location))
       location)))
 
+(defun bucket-region (bucket
+                      &key ((:credentials *credentials*) *credentials*))
+  (or (bucket-location bucket)
+      "us-east-1"))
+
+(defun region-endpoint (region)
+  (if (string= region "us-east-1")
+      "s3.amazonaws.com"
+      (format nil "s3-~A.amazonaws.com" region)))
+
 (defun query-bucket (bucket &key prefix marker max-keys delimiter
                      ((:credentials *credentials*) *credentials*))
   (submit-request (make-instance 'request
@@ -153,7 +163,7 @@ constraint."
          (bucket (bucket request)))
     (prog1
         (submit-request request)
-      (setf (redirected-endpoint endpoint bucket) nil))))
+      (setf (redirection-data endpoint bucket) nil))))
 
 
 ;;; Getting objects as vectors, strings, or files
@@ -169,7 +179,7 @@ constraint."
 (defun make-file-writer-handler (file &key (if-exists :supersede))
   (lambda (response)
     (check-request-success response)
-    (with-open-stream (input (body response))
+    (let ((input (body response)))
       (with-open-file (output file :direction :output
                                    :if-exists if-exists
                                    :element-type '(unsigned-byte 8))
@@ -181,7 +191,7 @@ constraint."
   (check-request-success response)
   (let ((buffer (make-octet-vector (content-length response))))
     (setf (body response)
-          (with-open-stream (input (body response))
+          (let ((input (body response)))
             (read-sequence buffer input)
             buffer))
     response))
@@ -201,15 +211,15 @@ constraint."
 
 
 (defun get-object (bucket key &key
-                   when-modified-since
-                   unless-modified-since
-                   when-etag-matches
-                   unless-etag-matches
-                   start end
-                   (output :vector)
-                   (if-exists :supersede)
-                   (string-external-format :utf-8)
-                   ((:credentials *credentials*) *credentials*))
+                                when-modified-since
+                                unless-modified-since
+                                when-etag-matches
+                                unless-etag-matches
+                                start end
+                                (output :vector)
+                                (if-exists :supersede)
+                                (string-external-format :utf-8)
+                                ((:credentials *credentials*) *credentials*))
   (flet ((range-argument (start end)
            (when start
              (format nil "bytes=~D-~@[~D~]" start (and end (1- end)))))
@@ -219,37 +229,46 @@ constraint."
       (setf start 0))
     (when (and start end (<= end start))
       (error "START must be less than END."))
-    (let ((request (make-instance 'request
-                                  :method :get
-                                  :bucket bucket
-                                  :key key
-                                  :extra-http-headers
-                                  (parameters-alist
-                                   :connection "close"
-                                   :if-modified-since
-                                   (maybe-date when-modified-since)
-                                   :if-unmodified-since
-                                   (maybe-date unless-modified-since)
-                                   :if-match when-etag-matches
-                                   :if-none-match unless-etag-matches
-                                   :range (range-argument start end))))
-          (handler (cond ((eql output :vector)
-                          'vector-writer-handler)
-                         ((eql output :string)
-                          (make-string-writer-handler string-external-format))
-                         ((eql output :stream)
-                          'stream-identity-handler)
-                         ((or (stringp output)
-                              (pathnamep output))
-                          (make-file-writer-handler output :if-exists if-exists))
-                         (t
-                          (error "Unknown ~S option ~S -- should be ~
+    (let* ((security-token (security-token *credentials*))
+           (request (make-instance 'request
+                                   :method :get
+                                   :bucket bucket
+                                   :key key
+                                   :amz-headers
+                                   (when security-token
+                                     (list (cons "security-token" security-token)))
+                                   :extra-http-headers
+                                   (parameters-alist
+                                    ;; nlevine 2016-06-15 -- not only is this apparently
+                                    ;; unnecessary, it also sends "connection" in the
+                                    ;; signed headers, which results in a
+                                    ;; SignatureDoesNotMatch error.
+                                    ;;  :connection (unless *use-keep-alive* "close")
+                                    :if-modified-since
+                                    (maybe-date when-modified-since)
+                                    :if-unmodified-since
+                                    (maybe-date unless-modified-since)
+                                    :if-match when-etag-matches
+                                    :if-none-match unless-etag-matches
+                                    :range (range-argument start end))))
+           (handler (cond ((eql output :vector)
+                           'vector-writer-handler)
+                          ((eql output :string)
+                           (make-string-writer-handler string-external-format))
+                          ((eql output :stream)
+                           'stream-identity-handler)
+                          ((or (stringp output)
+                               (pathnamep output))
+                           (make-file-writer-handler output :if-exists if-exists))
+                          (t
+                           (error "Unknown ~S option ~S -- should be ~
                                   :VECTOR, :STRING, :STREAM, or a pathname"
                                  :output output)))))
       (catch 'not-modified
         (handler-case
             (let ((response (submit-request request
-                                            :keep-stream (eql output :stream)
+                                            :keep-stream (or (eql output :stream)
+                                                             *use-keep-alive*)
                                             :body-stream t
                                             :handler handler)))
               (values (body response) (http-headers response)))
@@ -313,17 +332,17 @@ constraint."
 
 
 (defun put-object (object bucket key &key
-                   access-policy
-                   public
-                   metadata
-                   (string-external-format :utf-8)
-                   cache-control
-                   content-encoding
-                   content-disposition
-                   expires
-                   content-type
-                   (storage-class "STANDARD")
-                   ((:credentials *credentials*) *credentials*))
+                                       access-policy
+                                       public
+                                       metadata
+                                       (string-external-format :utf-8)
+                                       cache-control
+                                       content-encoding
+                                       content-disposition
+                                       expires
+                                       content-type
+                                       (storage-class "STANDARD")
+                                       ((:credentials *credentials*) *credentials*))
   (let ((content
          (etypecase object
            (string
@@ -332,7 +351,8 @@ constraint."
                                             string-external-format))
            ((or vector pathname) object)))
         (content-length t)
-        (policy-header (access-policy-header access-policy public)))
+        (policy-header (access-policy-header access-policy public))
+        (security-token (security-token *credentials*)))
     (setf storage-class (or storage-class "STANDARD"))
     (submit-request (make-instance 'request
                                    :method :put
@@ -341,8 +361,8 @@ constraint."
                                    :metadata metadata
                                    :amz-headers
                                    (append policy-header
-                                           (list (cons "storage-class"
-                                                       storage-class)))
+                                           (when security-token
+                                             (list (cons "security-token" security-token))))
                                    :extra-http-headers
                                    (parameters-alist
                                     :cache-control cache-control
@@ -465,12 +485,16 @@ constraint."
 ;;; Delete & copy objects
 
 (defun delete-object (bucket key &key
-                      ((:credentials *credentials*) *credentials*))
+                        ((:credentials *credentials*) *credentials*))
   "Delete one object from BUCKET identified by KEY."
-  (submit-request (make-instance 'request
-                                 :method :delete
-                                 :bucket bucket
-                                 :key key)))
+  (let ((security-token (security-token *credentials*)))
+    (submit-request (make-instance 'request
+                                   :method :delete
+                                   :bucket bucket
+                                   :key key
+                                   :amz-headers
+                                   (when security-token
+                                     (list (cons "security-token" security-token)))))))
 
 (defun bulk-delete-document (keys)
   (coerce
@@ -793,35 +817,67 @@ TARGET-BUCKET with a key prefix of TARGET-PREFIX."
              ssl bucket (url-encode key) sub-resource))
     ((nil)
      (format nil "http~@[s~*~]://s3.amazonaws.com/~@[~A/~]~@[~A~]~@[?~A~]"
-             ssl (url-encode bucket) (url-encode key) sub-resource))))
+             ssl
+             (url-encode bucket)
+             (url-encode key :encode-slash nil)
+             sub-resource))))
 
 (defun authorized-url (&key bucket key vhost expires ssl sub-resource
-                       ((:credentials *credentials*) *credentials*))
+                         ((:credentials *credentials*) *credentials*))
   (unless (and expires (integerp expires) (plusp expires))
     (error "~S option must be a positive integer" :expires))
-  (let* ((request (make-instance 'url-based-request
+  (let* ((region (bucket-region bucket))
+         (region-endpoint (region-endpoint region))
+         (endpoint (case vhost
+                     (:cname bucket)
+                     (:amazon (format nil "~A.~A" bucket region-endpoint))
+                     ((nil) region-endpoint)))
+         (request (make-instance 'url-based-request
                                  :method :get
                                  :bucket bucket
+                                 :region region
+                                 :endpoint endpoint
                                  :sub-resource sub-resource
                                  :key key
-                                 :expires (unix-time expires)))
-         (parameters
-          (alist-to-url-encoded-string
-           (list (cons "AWSAccessKeyId" (access-key *credentials*))
-                 (cons "Expires" (format nil "~D" (expires request)))
-                 (cons "Signature"
-                       (signature request))))))
-    (case vhost
-      (:cname
-       (format nil "http~@[s~*~]://~A/~@[~A~]?~@[~A&~]~A"
-               ssl bucket (url-encode key) sub-resource parameters))
-      (:amazon
-       (format nil "http~@[s~*~]://~A.s3.amazonaws.com/~@[~A~]?~@[~A&~]~A"
-               ssl bucket (url-encode key) sub-resource parameters))
-      ((nil)
-       (format nil "http~@[s~*~]://s3.amazonaws.com/~@[~A/~]~@[~A~]?~@[~A&~]~A"
-               ssl (url-encode bucket) (url-encode key) sub-resource
-                   parameters)))))
+                                 :expires (unix-time expires))))
+    (setf (amz-headers request) nil)
+    (setf (parameters request)
+          (parameters-alist "X-Amz-Algorithm" "AWS4-HMAC-SHA256"
+                            "X-Amz-Credential"
+                            (format nil "~A/~A/~A/s3/aws4_request"
+                                    (access-key *credentials*)
+                                    (iso8601-basic-date-string (date request))
+                                    (region request))
+                            "X-Amz-Date" (iso8601-basic-timestamp-string (date request))
+                            "X-Amz-Expires" (- expires (get-universal-time))
+                            "X-Amz-SignedHeaders"
+                            (format nil "~{~A~^;~}" (signed-headers request))))
+    (push (cons "X-Amz-Signature" (request-signature request))
+          (parameters request))
+    (let ((parameters (alist-to-url-encoded-string (parameters request))))
+      (case vhost
+        (:cname
+         (format nil "http~@[s~*~]://~A/~@[~A~]?~@[~A&~]~A"
+                 ssl
+                 bucket
+                 (url-encode key :encode-slash nil)
+                 sub-resource
+                 parameters))
+        (:amazon
+         (format nil "http~@[s~*~]://~A/~@[~A~]?~@[~A&~]~A"
+                 ssl
+                 endpoint
+                 (url-encode key :encode-slash nil)
+                 sub-resource
+                 parameters))
+        ((nil)
+         (format nil "http~@[s~*~]://~A/~@[~A/~]~@[~A~]?~@[~A&~]~A"
+                 ssl
+                 endpoint
+                 (url-encode bucket)
+                 (url-encode key :encode-slash nil)
+                 sub-resource
+                 parameters))))))
 
 
 ;;; Miscellaneous operations
